@@ -1,6 +1,6 @@
 /*
   Copyright (C) 2016  Brazil
-  Copyright (C) 2018-2022  Sutou Kouhei <kou@clear-code.com>
+  Copyright (C) 2018-2024  Sutou Kouhei <kou@clear-code.com>
 
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
@@ -18,8 +18,9 @@
 
 #include "../grn_ctx.hpp"
 #include "../grn_expr.h"
-#include "../grn_proc.h"
 #include "../grn_hash.h"
+#include "../grn_proc.h"
+#include "../grn_selector.h"
 #include "../grn_table_selector.h"
 
 #include <groonga/plugin.h>
@@ -164,6 +165,13 @@ namespace {
       }
     }
 
+    virtual bool
+    can_swap_result_set()
+    {
+      return selector_data_ &&
+             grn_selector_data_get_can_swap_result_set(ctx_, selector_data_);
+    }
+
     void
     init_table_selector(grn_ctx *ctx,
                         grn_table_selector *table_selector,
@@ -187,6 +195,11 @@ namespace {
           table_selector,
           max_n_enough_filtered_records_);
       }
+      if (can_swap_result_set()) {
+        grn_table_selector_set_ensure_using_select_result(ctx_,
+                                                          table_selector,
+                                                          true);
+      }
     }
 
     bool
@@ -476,11 +489,37 @@ namespace {
                 ctx->errline,
                 ctx->errbuf);
       }
+      auto base_result_set = *result_set;
+      const auto have_multiple_conditions =
+        (grn_expr_get_n_logical_ops(ctx, condition) >= 1);
+      // TODO: We can avoid this when condition uses only OR for
+      // logical operations.
+      const auto need_manual_or =
+        (op == GRN_OP_OR && have_multiple_conditions && base_result_set &&
+         grn_table_size(ctx, base_result_set) > 0);
+      if (need_manual_or) {
+        // We can't use *result_set as the base_result_set with this case.
+        // If we use it for 'A || query("...", "B C")', this expression
+        // is evaluated as '(A || B) && C' not 'A || (B && C)'.
+        base_result_set = nullptr;
+      }
       auto new_result_set =
-        grn_table_selector_select(ctx, &table_selector, *result_set);
+        grn_table_selector_select(ctx, &table_selector, base_result_set);
       if (new_result_set) {
         if (ctx->rc == GRN_SUCCESS) {
-          *result_set = new_result_set;
+          if (need_manual_or) {
+            grn_table_setoperation(ctx,
+                                   *result_set,
+                                   new_result_set,
+                                   *result_set,
+                                   op);
+            grn_obj_close(ctx, new_result_set);
+          } else {
+            *result_set = new_result_set;
+          }
+          if (can_swap_result_set()) {
+            grn_selector_data_set_result_set(ctx, selector_data_, *result_set);
+          }
         } else {
           if (new_result_set != *result_set) {
             grn_obj_close(ctx, new_result_set);
@@ -681,6 +720,13 @@ namespace {
       GRN_OBJ_FIN(ctx_, &sub_match_columns_);
     }
 
+  protected:
+    bool
+    can_swap_result_set() override
+    {
+      return false;
+    }
+
   private:
     bool
     prepare_match_columns_bulk() override
